/**
 * 
 */
package edx.jaz.schedule;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.HashMap;
import java.util.Random;

import javax.xml.transform.stream.StreamSource;

import org.jboss.logging.Logger;
import org.jboss.soa.esb.ConfigurationException;
import org.jboss.soa.esb.helpers.ConfigTree;
import org.jboss.soa.esb.listeners.ScheduledEventMessageComposer;
import org.jboss.soa.esb.listeners.message.MessageDeliverException;
import org.jboss.soa.esb.message.Message;
import org.jboss.soa.esb.message.MessagePayloadProxy;
import org.jboss.soa.esb.message.format.MessageFactory;
import org.jboss.soa.esb.schedule.SchedulingException;
import org.jboss.soa.esb.smooks.resource.SmooksResource;
import org.milyn.Smooks;
import org.milyn.SmooksException;
import org.milyn.container.ExecutionContext;
import org.milyn.event.report.HtmlReportGenerator;
import org.milyn.javabean.repository.BeanRepository;

import edx.jaz.RecordType;
import edx.jaz.Utilities;

/**
 * This class supports fetching files for ESB service processing.
 * The file fetch is triggered by the event scheduler.
 * The file content is not fetched, but the file is moved to working location 
 * and provides file://... URI reference for content fetch  
 * <br><br>
 * The following configuration parameters are required for FileFetchExecute trigger:<br>
 * <li> sourceDir = absolute path in the local filesystem, where file is ready for processing</li>
 * <li> suffix = suffix of the filename to be accepted for processing,<br>
 * the following filename pattern is accepted '<b>filename</b>'+'_'+'<b>suffix</b>' </li> 
 * <br/>    
 * --@throws ConfigurationException :       if parameters are invalid or incomplete<br>
 * --@throws RemoteFileSystemException :    if runtime error occurs
 * <br><br>
 *
 * @author edx
 *
 */
public class FileFetchExecute implements ScheduledEventMessageComposer {

	
	// configuration parameter names
	public static final String suffix = "suffix";
	public static final String smooksConfig = "smooksConfig";
	public static final String smooksReportPath = "reportPath";
	public static final String charset = "charset";
	
	// parameters evaluated from configuration
	private String sourcePath;
	private RecordType rectype;
    private Smooks smooks;
    private String reportPath;
    private String encoding;

    // other private objects
	private Random rand;
    private MessagePayloadProxy payloadProxy;
	
	private Logger log = Logger.getLogger(this.getClass().getSimpleName());
	
	/**
	 * 
	 */
	public FileFetchExecute() {
		// empty constructor, all initialization is in initialize() method
	}

	/* (non-Javadoc)
	 * @see org.jboss.soa.esb.listeners.ScheduledEventMessageComposer#composeMessage()
	 */
	public Message composeMessage() throws SchedulingException {
		// the algorithm for preparing the message is:
		// get the list of files in sourcePath
		// empty ? then return null
		// not empty ? pick first
		//             rename to filename_batchid_WORK
		//             construct URI and insert it to RelatesTO header field
		//             add info message about the file to the message payload
		// batchid is a hexadecimal representation of the random INT value (32bit).
		
		File srcDir = new File(sourcePath);
		File[] flist = srcDir.listFiles(new InputFileFilter(rectype.toString()));
		if( flist == null || flist.length == 0 ) {
			log.info("no files in "+srcDir.getAbsolutePath()+" of the type "+rectype+" to process");
			return null;
		}
		
		// take flist[0] for processing

		// evaluate pick file, rename it to prevent from processing by other 
		File pick = new File(flist[0].getAbsolutePath()+"_WORK");
		// rename it and store URI to RelatesTo Header
		flist[0].renameTo(pick);
		log.info("File '"+pick.getPath()+"' prepared for processing.");			
		
		// try to recognize number of records for processing
		Long nr = new Long(0);
		try {
			String wcinfo = "";
			String line;
			Process p = Runtime.getRuntime().exec("wc "+pick.getAbsolutePath());
			BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
			while ((line = input.readLine()) != null) {
		        wcinfo += line;
		      }
			input.close();
			log.info("WC INFO: "+wcinfo);
			nr = Long.valueOf(wcinfo.trim().split(" +",2)[0])-1L; // we skip the first line in record counts
		} catch (IOException e1) {
			log.warn("Cannot determine WC characteristic of the file "+pick.getName());
		}

		// create unique batchID for context identification
		String batchid = Integer.toHexString(rand.nextInt()); 
		
		// Prepare Smooks ExecutionContext and add Batch Info bean to the BeanRepository
        ExecutionContext executionContext = smooks.createExecutionContext();

    	BeanRepository beanRepo = BeanRepository.getInstance(executionContext);
    	beanRepo.addBean(Utilities.batchId,batchid);
    	beanRepo.addBean(Utilities.recType,rectype);        	
        
        if(reportPath != null) {
            try {
                executionContext.setEventListener(new HtmlReportGenerator(reportPath+"/"+batchid+".html"));
            } catch (IOException e) {
            	log.warn("IOException: "+e.getMessage());
                throw new SchedulingException("Failed to create HtmlReportGenerator instance.", e);
            }
        }

        // and finally run Smooks filtering
        Date startAt = new Date();
		try {
			InputStream is = new FileInputStream(pick);
			InputStreamReader isr = new InputStreamReader( is, encoding );
			smooks.filterSource(executionContext,new StreamSource( isr ));
		} catch (SmooksException e) {
			// we do not throw away if smooks finishes with errors 
			log.warn("Error processing file, SmooksException: "+e.getMessage());
		} catch (FileNotFoundException f) {
			throw new SchedulingException("FileNotFoundException: "+f.getMessage());
		} catch (UnsupportedEncodingException e) {
			throw new SchedulingException("Cannot read file, UnsupportedEncodingException:"+e.getMessage() );
		} 
		// And now gather info from result and store into resulting message
		Message message = MessageFactory.getInstance().getMessage();
		
		HashMap<String,Serializable> hm = new HashMap<String,Serializable>();
		hm.put(Utilities.sourcePath, sourcePath);
		hm.put(Utilities.fileName, flist[0].getName());
		hm.put(Utilities.batchId, batchid);
		hm.put(Utilities.recType, rectype);
		hm.put(Utilities.totalRecords, nr);
		Long readrec = (Long) beanRepo.getBean("counter");
		hm.put(Utilities.recordCounter, (readrec==null)?(Long.valueOf(0)):(readrec));
		hm.put(Utilities.readTime, Long.valueOf(System.currentTimeMillis()-startAt.getTime()) );
		
        try {
            payloadProxy.setPayload( message, hm );
        } catch (MessageDeliverException e) {
            throw new SchedulingException("MessgeDeliveryException while trying to retrieve the message payload:", e);
        }

        return message;

	}

	/* (non-Javadoc)
	 * @see org.jboss.soa.esb.listeners.ScheduledEventMessageComposer#onProcessingComplete(org.jboss.soa.esb.message.Message)
	 */
	public Message onProcessingComplete(Message message)
			throws SchedulingException {
		// nothing to do
		return message;
	}

	/* (non-Javadoc)
	 * @see org.jboss.soa.esb.Initializable#initialize(org.jboss.soa.esb.helpers.ConfigTree)
	 */
	public void initialize(ConfigTree config) throws ConfigurationException {

		// check existence and create if missing the source directory
		sourcePath = config.getAttribute(Utilities.sourcePath, "/usr/local/jboss/run/input");
		File srcDir = new File(sourcePath);
		try {
			if( !srcDir.exists() )
				if( !srcDir.mkdirs() )
					throw new ConfigurationException("Cannot create directory: "+sourcePath);
		} catch (SecurityException se ) {
			throw new ConfigurationException("Cannot create directory '"+sourcePath+"', SecurityException: "+se.getMessage());
		}
		
		rectype = RecordType.valueOf( config.getRequiredAttribute(suffix) );
		// validate against RecordType
		if( rectype == null )
			throw new ConfigurationException("Submitted suffix does not indicate Record Type, must be one of: "+RecordType.values() );

		// initialize random number generator 
		rand = new Random( System.currentTimeMillis() );
		
		payloadProxy = new MessagePayloadProxy(config);
		
		// initialize smooks engine
		final String smooksConfiguration = config.getRequiredAttribute(smooksConfig);
        try
        {
            smooks = SmooksResource.createSmooksResource(smooksConfiguration);
        }
        catch (Exception e) {
            throw new ConfigurationException("Failed to create Smooks instance for config '" + smooksConfig + "'.", e);
        }
        
        // create reportPath for debugging
        reportPath = config.getAttribute(smooksReportPath);
        
        // and encoding of stream when reading source data files
        encoding = config.getAttribute(charset, "UTF-8");
        
		log.info("initialized "+Utilities.sourcePath+"="+sourcePath
				+" "+Utilities.recType+"="+rectype
				+" "+charset+"="+encoding 
				+" "+smooksConfig+"="+smooksConfiguration
				+" "+smooksReportPath+"="+((reportPath==null)?"(NULL)":reportPath) );
	}

	/* (non-Javadoc)
	 * @see org.jboss.soa.esb.Initializable#uninitialize()
	 */
	public void uninitialize() {
        SmooksResource.closeSmooksResource(smooks);
 	}

	/**
	 * Filter class to accept filenames of the given suffix.
	 */
	private class InputFileFilter implements FilenameFilter {

		private final String sfx;
		
		public InputFileFilter(String suffix) {
			sfx = (suffix!=null)?("_"+suffix):null;
		}
		
		public boolean accept(File dir, String name) {
			if( sfx!=null ) {
				File ff = new File(dir,name);
				if( ff.isFile() && name.endsWith(sfx) )
					return true;
			}
			return false;
		}
		
	}
	
}
